/*
 * gsampler-database.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
 * 
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <glib/gi18n.h>

#include "gsampler-window.h"
#include "gsampler-database.h"
#include "gsampler-utils.h"
#include "gsampler-channel-manager.h"
#include "gsampler-client-database.h"
#include "gsampler-client-channel.h"
#include "gsampler-database-search.h"
#include "gsampler-database-instinfo.h"

#if GTK_CHECK_VERSION (2,90,0)
#define DB_ICON "db.svg"
#else
#define DB_ICON "db.png"
#endif

G_DEFINE_TYPE (GsamplerDataBase, gsampler_data_base, GTK_TYPE_BOX);

static GHashTable* gsampler_data_base_path_list = NULL;
static GSList *path_list = NULL;

static void gsampler_data_base_update_instrument_treeview (GsamplerDataBase* self);

static gchar* 
gsampler_data_base_get_parent_path (GsamplerDataBase* self, const gchar* dir, gchar** rep) {
	gchar* base = NULL;
	gchar* result = NULL;
	gchar* temp;
	gchar** places;
	gint places_length = 0;
	GString* builder;
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (dir != NULL, NULL);

	if (g_strcmp0 (dir, "/") == 0)
		return NULL;
	temp = g_strdup (dir);
	if (g_str_has_prefix (temp, "/")) {
		temp = string_slice (temp, 1, strlen (temp));
	}
	if (g_str_has_suffix (temp, "/")) {
		temp = string_slice (temp, 0, (strlen (temp) - 1));
	}
	places = g_strsplit (temp, "/", 0);
	if (places != NULL)
		places_length = g_strv_length (places);
	builder = g_string_new ("");
	if (places[1] != NULL) {
		int i = -1;
		int place_it;
		for (place_it = 0; place_it < places_length; place_it++) {
			i++;
		}
		int y;
		for (y = 0; y < i; y++) {
			g_string_append (builder, g_strconcat ("/", places[y], NULL));
		}
		base = g_strdup (places[i]);
	} else {
		base = g_strdup (places[0]);
	}
	result = g_strdup (builder->str);
	g_string_free (builder, TRUE);
	g_strfreev (places);
	g_free (temp);
	if (rep) {
		*rep = base;
	} else {
		g_free (base);
	}
	return result;
}

static gchar*
gsampler_data_base_get_full_path (GsamplerDataBase *self, GtkTreeModel *model, GtkTreeIter *iter)
{
	gchar* tmp_name = NULL;
	GString* builder = NULL;
	GtkTreePath* path = NULL;
	gchar *result = NULL;
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (model != NULL, NULL);
	g_return_val_if_fail (iter != NULL, NULL);
	
	builder = g_string_new ("");
	path = gtk_tree_model_get_path (model, iter);
	if (path != NULL)
	{
		while (gtk_tree_path_get_depth (path) > 0)
		{
			gtk_tree_model_get_iter (model, iter, path);
			gtk_tree_model_get (model, iter, 1, &tmp_name, -1);
			if (tmp_name != NULL) {
				if (g_str_has_suffix (tmp_name, "/"))
					g_string_prepend (builder, tmp_name);
				else
					g_string_prepend (builder, g_strconcat (tmp_name, "/", NULL));
			}			    
			gtk_tree_path_up (path);
		}
		result = g_strdup (builder->str);
		g_free (tmp_name);
		gtk_tree_path_free (path);
	}
	g_string_free (builder, TRUE);
	
	return result;
}

static gchar*
gsampler_data_base_get_db_path (GsamplerDataBase *self)
{
	GtkTreeModel *model = NULL;
	GtkTreeSelection *select = NULL;
	GtkTreeIter iter = {0};
	gchar *result = NULL;
	g_return_val_if_fail (self != NULL, NULL);

	select = gtk_tree_view_get_selection (self->priv->db_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
		result = gsampler_data_base_get_full_path (self, model, &iter);
	return result;	
}

static gchar*
gsampler_data_base_get_inst_name (GsamplerDataBase *self)
{
	GtkTreeModel *model = NULL;
	GtkTreeSelection *select = NULL;
	GtkTreeIter iter = {0};
	gchar *result = NULL;
	g_return_val_if_fail (self != NULL, NULL);
	
	select = gtk_tree_view_get_selection (self->priv->inst_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter))
		gtk_tree_model_get (model, &iter, 1, &result, -1);
	return result;
}

static gchar*
gsampler_data_base_get_inst_path (GsamplerDataBase *self)
{
	gchar *dir = NULL;
	gchar *name = NULL;
	gchar *result = NULL;
	g_return_val_if_fail (self != NULL, NULL);

	dir = gsampler_data_base_get_db_path (self);
	name = gsampler_data_base_get_inst_name (self);
	if (dir != NULL && name != NULL)
		result = g_strconcat (dir, name, NULL);
	g_free (name);
	g_free (dir);
	
	return result;
}

static gboolean
gsampler_data_base_check_selected_is_dir (GsamplerDataBase *self)
{
	GtkTreeModel* model = NULL;
	GtkTreeIter iter = {0};
	GtkTreeSelection* select = NULL;
	gboolean is_dir = FALSE;
	g_return_val_if_fail (self != NULL, FALSE);
	
	select = gtk_tree_view_get_selection (self->priv->inst_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
	{
		gtk_tree_model_get (model, &iter, 2, &is_dir, -1);
	}
	return is_dir;
}

static void 
gsampler_data_base_update_instrument_treeview (GsamplerDataBase* self)
{
	GtkTreeIter iter = {0};
	GtkListStore* list = NULL;
	GsamplerClient* client = NULL;
	gchar* path;
	gchar** dir_list = NULL;
	gchar** inst_list = NULL;
	gint dir_list_length = 0;
	gint inst_list_length = 0;
	gint it;
	gchar* inst = NULL;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();

	path = gsampler_data_base_get_db_path (self);
	if (path == NULL) return;

	list = (GtkListStore*) gtk_tree_view_get_model (self->priv->inst_tree);
	gtk_list_store_clear (list);
	
	dir_list = gsampler_client_db_list_dir (client, path);
	if (dir_list != NULL) 	dir_list_length = g_strv_length (dir_list);
	for (it = 0; it < dir_list_length; it++) 
	{
		inst = string_replace (dir_list[it], "'", "");
		if (g_strcmp0 (inst, "NONE") != 0) 
		{
			gtk_list_store_append (list, &iter);
			gtk_list_store_set (list, &iter, 1, inst, 2, TRUE, -1);
		}
		g_free (inst);
	}

	inst_list = gsampler_client_db_list_instrument (client, path);
	if (inst_list != NULL) 	inst_list_length = g_strv_length (inst_list);
	for (it = 0; it < inst_list_length; it++) 
	{
		inst = string_replace (inst_list[it], "'", "");
		if (g_strcmp0 (inst, "NONE") != 0) 
		{
			gtk_list_store_append (list, &iter);
			gtk_list_store_set (list, &iter, 1, inst, 2, FALSE, -1);
		}
		g_free (inst);
	}
	g_free (path);
	g_strfreev (inst_list);
	g_object_unref (client);
}

static void 
gsampler_data_base_on_folder_changed (GtkTreeSelection* select, gpointer data) 
{
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (select != NULL);
	
	gsampler_data_base_update_instrument_treeview (self);
}

static void 
gsampler_data_base_add_db_store (GsamplerDataBase* self, const gchar* dir) 
{
	GtkTreeIter iter = {0};
	GtkTreePath* path = NULL;
	gchar* rep = NULL;
	GtkTreeStore* store;
	gchar* parent_dir = NULL;
	g_return_if_fail (self != NULL);
	g_return_if_fail (dir != NULL);
	
	store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->db_tree);
	parent_dir = gsampler_data_base_get_parent_path (self, dir, &rep);
	if (parent_dir != NULL)
	{	
		path = g_hash_table_lookup (gsampler_data_base_path_list, parent_dir);
	} else 
	{
		path = g_hash_table_lookup (gsampler_data_base_path_list, dir);
	}
	if (path != NULL) 
	{
		GtkTreeIter parent = {0};
		gtk_tree_model_get_iter ((GtkTreeModel*) store, &parent, path);
		gtk_tree_store_append (store, &iter, &parent);
	} else 
	{
		gtk_tree_store_append (store, &iter, NULL);
	}
	if (rep != NULL)
	{
		gtk_tree_store_set (store, &iter, 1, rep, -1);
	} else
	{
		gtk_tree_store_set (store, &iter, 1, dir, -1);
	}
	path = gtk_tree_model_get_path ((GtkTreeModel*) store, &iter);
	g_hash_table_insert (gsampler_data_base_path_list, g_strdup (dir), gtk_tree_path_copy (path));
	g_free (parent_dir);
	g_free (rep);
	gtk_tree_path_free (path);
}

static void 
gsampler_data_base_init_data_to_store (GsamplerDataBase* self, const gchar* orig, 
                                       gchar** reps, GtkTreeIter* parent) 
{
	GtkTreeStore* store = NULL;
	GtkTreePath* path = NULL;
	GtkTreeIter child_iter = {0};
	GtkTreeIter parent_iter = {0};
	gchar* rep = NULL;
	gchar* new_orig = NULL;
	gchar** new_rep = NULL;
	GsamplerClient* client = NULL;
	gint rep_it;
	gint reps_lenght = 0;
	g_return_if_fail (self != NULL);
	g_return_if_fail (orig != NULL);
	g_return_if_fail (reps != NULL);
	
	client = gsampler_client_get_instance ();
	store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->db_tree);
	if (parent == NULL)
	{
		path = g_hash_table_lookup (gsampler_data_base_path_list, "/");
		if (path != NULL)
		{
			gtk_tree_model_get_iter ((GtkTreeModel*) store, &parent_iter, path);
		}
	} else
	{
		parent_iter = (*parent);
	}

	if (reps != NULL)
		reps_lenght = g_strv_length (reps);
	for (rep_it = 0; rep_it < reps_lenght; rep_it++) 
	{
		rep = string_replace (reps[rep_it], "'", "");
		gtk_tree_store_append (store, &child_iter, &parent_iter);
		gtk_tree_store_set (store, &child_iter, 1, rep, -1);
		path = gtk_tree_model_get_path ((GtkTreeModel*) store, &child_iter);
		if (g_str_has_prefix (rep, "/") || g_strcmp0 (orig, "/") == 0)
			new_orig = g_strconcat (orig, rep, NULL);
		else
			new_orig = g_strconcat (orig, "/", rep, NULL);
		g_hash_table_insert (gsampler_data_base_path_list, 
		                     g_strdup (new_orig), gtk_tree_path_copy (path));
		new_rep = gsampler_client_db_list_dir (client, new_orig);
		if (new_rep != NULL && strlen (new_rep[0]) != 0) 
		{
			gsampler_data_base_init_data_to_store (self, new_orig, new_rep, &child_iter);
		}
		g_strfreev (new_rep);
		g_free (new_orig);
		g_free (rep);
		gtk_tree_path_free (path);
	}
	g_object_unref (client);
}

static gboolean 
gsampler_data_base_update_dir_treeview (GsamplerDataBase* self) 
{
	GsamplerClient * client = NULL;
	gchar **db_base = NULL;
	GtkTreeStore* store = NULL;
	GtkTreeIter sel_iter = {0};
	GtkTreePath *sel_path = NULL;
	GtkTreePath *path = NULL;
	GtkTreeSelection* select = NULL;
	g_return_val_if_fail (self != NULL, FALSE);

	select = gtk_tree_view_get_selection (self->priv->db_tree);
	gtk_tree_selection_get_selected (select, NULL, &sel_iter);
	store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->db_tree);
	if (gtk_tree_store_iter_is_valid (store, &sel_iter))
		sel_path = gtk_tree_model_get_path ((GtkTreeModel*) store, &sel_iter);
	gtk_tree_store_clear (store);
	g_hash_table_remove_all (gsampler_data_base_path_list);
	
	client = gsampler_client_get_instance ();
	db_base = gsampler_client_db_list_dir (client, "/");
	gsampler_data_base_add_db_store (self, "/");
	if (db_base != NULL && (strlen (db_base[0]) != 0 || g_strcmp0 (db_base[0], "NONE") != 0)) 
	{
		gsampler_data_base_init_data_to_store (self, "/", db_base, NULL);
		path = g_hash_table_lookup (gsampler_data_base_path_list, "/");
		if (path != NULL)
		{
			gtk_tree_view_expand_row (self->priv->db_tree, path, FALSE);
		}
		g_strfreev (db_base);
	} else
	{
		g_warning ("Can't update DB view, database is empty");
		return FALSE;
	}
	g_object_unref (client);
	if (sel_path != NULL)
		gtk_tree_selection_select_path (select, sel_path);
	gtk_tree_path_free (sel_path);
	return TRUE;
}

static gboolean 
gsampler_data_base_update_dir_treeview_func (gpointer data)
{
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (self != NULL, FALSE);

	gsampler_data_base_update_dir_treeview (self);
	return FALSE;
}

static void 
gsampler_data_base_on_db_path_count (GsamplerClient* client, const gchar* path, gpointer data) 
{
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (path != NULL);

	self->priv->new_rep = g_strdup (path);
	gdk_threads_add_idle_full (G_PRIORITY_DEFAULT_IDLE, gsampler_data_base_update_dir_treeview_func,
	                 		   g_object_ref (self), g_object_unref);
} 

static gboolean 
gsampler_data_base_initialyse (gpointer data) 
{
	GsamplerClient* client = NULL;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (GSAMPLER_IS_DATA_BASE (self), FALSE);

	if (!g_source_is_destroyed (g_main_current_source ()))
	{
		client = gsampler_client_get_instance ();
		if (!gsampler_data_base_update_dir_treeview (self)) 
		{
			g_warning ("Can't initialize DB, database is empty");
		}
		//TODO
		gtk_widget_set_sensitive ((GtkWidget*) self, TRUE);
		self->priv->coun_sig_id = g_signal_connect_object (client, "db-path-count-event", 
					                 					   (GCallback) gsampler_data_base_on_db_path_count,
		                                                   self, 0);
		g_object_unref (client);
	}	
	return FALSE;
}

static gboolean 
gsampler_data_base_add_db_instrument (gpointer data) {
	GsamplerClient* client = NULL;
	GsamplerDbJobInfo* info = NULL;
	gdouble perc = 0;
	gchar* text = NULL;
	gboolean result = FALSE;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (self != NULL, FALSE);

  	if (!g_source_is_destroyed (g_main_current_source ()))
    {
		if (self->priv->add_job_id <= 0)
			return FALSE;
	
		client = gsampler_client_get_instance ();
		info = gsampler_client_db_get_job_info (client, self->priv->add_job_id);
		if (info->scanned == 0 && info->total !=0) 
		{
			result = TRUE;
		} else if (info->scanned < info->total) 
		{
			perc = (gdouble) (info->scanned / info->total);
			text = g_strdup_printf ("Scanning %.0f of %.0f files", info->scanned, info->total);
						gsampler_database_progress_update (self->priv->db_progress, perc, text);
			g_free (text);
			result = TRUE;
			//gdk_flush ();
			//while (gtk_events_pending ()) {	gtk_main_iteration ();}
		} else {
			perc = (gdouble) (info->scanned / info->total);
			text = g_strdup_printf ("%.0f of %.0f files scanned", info->scanned, info->total);
			gsampler_database_progress_update (self->priv->db_progress, perc, text);
			g_free (text);
			self->priv->coun_sig_id = g_signal_connect_object (client, "db-path-count-event", 
					                 					   (GCallback) gsampler_data_base_on_db_path_count,
		                                                   self, 0);
			self->priv->add_job_id = 0;
			result = FALSE;
		}

		gsampler_db_job_info_destroy (info);
		g_object_unref (client);
	}
	return result;
}

static void 
gsampler_data_base_on_add_db_clicked (GtkToolButton* button, gpointer data) 
{
	GsamplerMainWindow* win = NULL;
	GsamplerOption* option = NULL;
	GtkDialog* dialog = NULL;
	GtkBox* vbox1 = NULL;
	GtkEntry* db_chooser = NULL;
	GtkFileChooserButton* dir_chooser = NULL;
	gchar* path = NULL;
	GtkLabel* db_label = NULL;
	GtkLabel* dir_label = NULL;
	GtkHBox* hbox1 = NULL;
	GtkHBox* hbox2 = NULL;
	gint result;
	gchar *db_path = NULL;
	gchar *db_base = NULL;
	GsamplerClient* client = NULL;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (button != NULL);

	win = gsampler_main_window_get_instance ();
	option = gsampler_option_get_instance ();
	dialog = (GtkDialog*) gtk_dialog_new_with_buttons (_("Add to DB"), (GtkWindow*) win,
	                                                   GTK_DIALOG_MODAL, 
	                                                   GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
	                                                   GTK_STOCK_APPLY, GTK_RESPONSE_APPLY, NULL);
	vbox1 = (GtkBox*) gtk_dialog_get_content_area (dialog);;
	db_chooser = (GtkEntry*) gtk_entry_new ();
	gtk_entry_set_text (db_chooser, "/");
	dir_chooser = (GtkFileChooserButton*) gtk_file_chooser_button_new ("Select Samples dir", 
	                                                                   GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
	path = gsampler_option_get_string (option, "Client", "sample-path");
	gtk_file_chooser_set_current_folder ((GtkFileChooser*) dir_chooser, path);
	db_label = (GtkLabel*) gtk_label_new (_("Select db dir : "));
	dir_label = (GtkLabel*) gtk_label_new (_("Select system dir : "));
	hbox1 = (GtkHBox*) gtk_hbox_new (TRUE, 5);
	hbox2 = (GtkHBox*) gtk_hbox_new (TRUE, 5);
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) dir_label, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) dir_chooser, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) hbox2, (GtkWidget*) db_label, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) hbox2, (GtkWidget*) db_chooser, FALSE, TRUE, 0);
	gtk_box_pack_start (vbox1, GTK_WIDGET (hbox1), FALSE, TRUE, 0);
	gtk_box_pack_start (vbox1, GTK_WIDGET (hbox2), FALSE, TRUE, 0);
	gtk_widget_show_all ((GtkWidget*) dialog);
	result = gtk_dialog_run (dialog);
	if (result == GTK_RESPONSE_APPLY) 
	{
		self->priv->db_progress = gsampler_database_progress_new ();
		gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) self->priv->db_progress, FALSE, TRUE, 0);
		gtk_box_reorder_child ((GtkBox*) self, (GtkWidget*) self->priv->db_progress, 0);
#ifdef G_OS_WIN32
		db_path = string_replace (gtk_file_chooser_get_filename ((GtkFileChooser*) dir_chooser), "\\", "/");
#else
		db_path = gtk_file_chooser_get_filename ((GtkFileChooser*) dir_chooser);
#endif
		db_base = g_strdup((gchar*) gtk_entry_get_text (db_chooser));
		client = gsampler_client_get_instance ();
		self->priv->add_job_id = gsampler_client_db_add_instrument (client, db_base, db_path, 
		                                                            TRUE, FALSE, FALSE, -1);
		if (self->priv->add_job_id > 0) 
		{
			g_signal_handler_disconnect (client, self->priv->coun_sig_id);
			/*g_signal_handlers_disconnect_by_func (client, gsampler_data_base_on_db_path_count, NULL);*/
			gdk_threads_add_timeout_seconds (1, gsampler_data_base_add_db_instrument, self);
		} else 
		{
			g_warning ("Can't add db instrument job_id : %d", self->priv->add_job_id);
		}
		g_object_unref (client);
	}
	gtk_widget_destroy ((GtkWidget*) dialog);
	g_free (path);
	g_object_unref (option);
	g_object_unref (win);
}

static void 
gsampler_data_base_on_clear_db_clicked (GtkToolButton* button, gpointer data)
{
	GString *ret = NULL;
	gchar **files = NULL;
	GsamplerClient* client;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (GSAMPLER_IS_DATA_BASE (self));
	g_return_if_fail (button != NULL);

	client = gsampler_client_get_instance ();
	files = gsampler_client_db_check_lost_inst_files (client);
	if (files != NULL)
	{
		ret = g_string_new("");
		int i;
		for (i=0; i != g_strv_length (files); i++)
		{
			if (files[i] != NULL)
				g_string_append (ret, files[i]);
		}
		if (ret->str != NULL && g_strcmp0 (ret->str, "") != 0)
			gsampler_main_window_send_message ("info", ret->str);
		g_string_free (ret, TRUE);
		g_strfreev (files);
	}
	g_object_unref (client);
}

static gboolean 
gsampler_data_base_format_db_instrument (gpointer data) 
{
	GsamplerClient* client;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (self != NULL, FALSE);

  	if (!g_source_is_destroyed (g_main_current_source ()))
    {
		client = gsampler_client_get_instance ();
		if (!gsampler_client_db_format_instruments (client)) 
		{
			gsampler_main_window_send_message ("warning", "Can't format the database");
		} else 
		{
			GtkTreeStore* store = NULL;
			GtkListStore* liste = NULL;
	
			liste = (GtkListStore*) gtk_tree_view_get_model (self->priv->inst_tree);
			gtk_list_store_clear (liste);
			g_slist_free_full (path_list, g_free);
			g_hash_table_remove_all (gsampler_data_base_path_list);
			store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->db_tree);
			gtk_tree_store_clear (store);
		}
		g_object_unref (client);
	}
	return FALSE;
}


static void 
gsampler_data_base_on_format_db_clicked (GtkToolButton* button, gpointer data) 
{
	GsamplerMainWindow* win;
	GtkDialog* dialog;
	gint response;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (GSAMPLER_IS_DATA_BASE (self));
	g_return_if_fail (button != NULL);
	
	win = gsampler_main_window_get_instance ();
	dialog = (GtkDialog*) gtk_message_dialog_new ((GtkWindow*) win, GTK_DIALOG_MODAL, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK_CANCEL, "%s", _("This will format the DB are you sure?"));
	gtk_dialog_set_default_response ((GtkDialog*) dialog, (gint) GTK_RESPONSE_CANCEL);
	response = gtk_dialog_run (dialog);
	if (response == GTK_RESPONSE_OK) 
	{
		gdk_threads_add_idle_full (G_PRIORITY_HIGH_IDLE, gsampler_data_base_format_db_instrument, 
		                 		   g_object_ref (self), g_object_unref);
	}
	gtk_widget_destroy ((GtkWidget*) dialog);
	g_object_unref (win);
}

static void
gsampler_data_base_on_refresh_db_clicked (GtkToolButton* button, gpointer data) 
{
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (GSAMPLER_IS_DATA_BASE (self));
	g_return_if_fail (button != NULL);

	gsampler_data_base_update_dir_treeview (self);	
}

static void
gsampler_data_base_on_find_select_event (GsamplerDatabaseSearch* dialog, gchar* name, gpointer data)
{
	gchar* base;
	gchar* inst = NULL;
	GtkTreePath *path = NULL;
	GtkTreePath *up_path = NULL;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (GSAMPLER_IS_DATA_BASE (self));
	
	base = gsampler_data_base_get_parent_path (self, name, &inst);
	path = g_hash_table_lookup (gsampler_data_base_path_list, base);
	if (path != NULL)
	{
		gint depth = gtk_tree_path_get_depth (path);
		if (depth > 1) 
		{
			up_path = gtk_tree_path_copy (path);
			while (gtk_tree_path_up (up_path)) {
				gtk_tree_view_expand_row (self->priv->db_tree, up_path, TRUE);
			}
		}
		GtkTreeSelection *select = gtk_tree_view_get_selection (self->priv->db_tree);
		gtk_tree_selection_select_path (select, path);
	}
}

static void 
gsampler_data_base_on_find_db_clicked (GtkToolButton* button, gpointer data) 
{
	GsamplerDatabaseSearch* dialog;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (GSAMPLER_IS_DATA_BASE (self));
	g_return_if_fail (button != NULL);

	dialog = gsampler_database_search_new ();
	g_signal_connect_object (dialog, "gsampler-database-search-select",
	                         (GCallback) gsampler_data_base_on_find_select_event, self, 0);
	gtk_widget_show_all ((GtkWidget*) dialog);
	gtk_dialog_run ((GtkDialog*) dialog);
	gtk_widget_destroy ((GtkWidget*) dialog);
}

static gboolean 
gsampler_data_base_load_on (gpointer data) 
{
	GtkTreeIter iter = {0};
	GtkTreeModel* model = NULL;
	GtkListStore* list = NULL;
	GtkTreeSelection* select = NULL;
	gchar* tmp_name = NULL;
	GString* builder = NULL;
	GtkTreePath* path = NULL;
	GsamplerClient* client = NULL;
	GsamplerDbInstInfo *file = NULL;
	GsamplerChannelManager* chan_manag = NULL;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (self != NULL, FALSE);

  	if (!g_source_is_destroyed (g_main_current_source ()))
    {
		list = (GtkListStore*) gtk_tree_view_get_model (self->priv->inst_tree);
		if (!gtk_list_store_iter_is_valid (list, &self->priv->current_iter)) {
			g_warning ("file %s: line %d: error: Invalide Iter", __FILE__, __LINE__);
			return FALSE;
		}
		select = gtk_tree_view_get_selection (self->priv->db_tree);
		if (!gtk_tree_selection_get_selected (select, &model, &iter)) {
			g_warning ("file %s: line %d: error: No Selection", __FILE__, __LINE__);
			return FALSE;
		}
		builder = g_string_new ("");
		path = gtk_tree_model_get_path (model, &iter);
		do {
			gtk_tree_model_get_iter (model, &iter, path);
			gtk_tree_model_get (model, &iter, 1, &tmp_name, -1);
			tmp_name = g_strconcat ("/", tmp_name, NULL);
			g_string_prepend (builder, tmp_name);
			gtk_tree_path_up (path);
		} while (gtk_tree_path_get_depth (path) > 0);
		gtk_tree_model_get ((GtkTreeModel*) list, &self->priv->current_iter, 1, &tmp_name, -1);
		if (tmp_name == NULL) {
			g_warning ("file %s: line %d: error: No Tmp NAME", __FILE__, __LINE__);
			g_string_free (builder, TRUE);
			gtk_tree_path_free (path);
			return FALSE;
		}
		tmp_name = g_strconcat ("/", tmp_name, NULL);
		g_string_append (builder, tmp_name);
		g_free (tmp_name);
		client = gsampler_client_get_instance ();
		file = gsampler_client_db_get_instrument_info (client, builder->str);
		chan_manag = gsampler_channel_manager_get_instance ();
		if (self->priv->chan_id == (-1)) {
			self->priv->chan_id = gsampler_channel_manager_create_channel (chan_manag);
		}
		while (gtk_events_pending ()) { gtk_main_iteration (); }
		if (self->priv->chan_id != (-1))
			gsampler_channel_manager_update_channel (chan_manag, self->priv->chan_id, file->format_family, file->file, &file->number);
		g_object_unref (chan_manag);
		gsampler_db_inst_info_free (file);
		g_object_unref (client);
		gtk_tree_path_free (path);
		g_string_free (builder, TRUE);
	}
	return FALSE;
}

static void 
gsampler_data_base_load_on_chan (GtkMenuItem* item, gpointer data) 
{
	gchar* tmp = NULL;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);
	
	tmp = (gchar*) gtk_menu_item_get_label (item);
	tmp = string_replace (tmp, "Channel ", "");
	self->priv->chan_id = atoi (tmp);
	g_free (tmp);
	gdk_threads_add_idle_full (G_PRIORITY_DEFAULT_IDLE, gsampler_data_base_load_on,
	                 	  	   g_object_ref (self), g_object_unref);
}

static void 
gsampler_data_base_load_on_new (GtkMenuItem* item, gpointer data) 
{
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);
	
	self->priv->chan_id = -1;
	gdk_threads_add_idle_full (G_PRIORITY_DEFAULT_IDLE, gsampler_data_base_load_on,
	                 	  	   g_object_ref (self), g_object_unref);
}

static GtkWidget*
gsampler_data_base_tree_dialog (GsamplerDataBase *self, GtkWidget *tree_view)
{
	GsamplerMainWindow *parent_win = NULL;
	GtkWidget *dialog = NULL;
	GtkWidget *content_area = NULL;
	GtkTreeModel *db_model = NULL;
	GtkTreeIter iter = {0};
	GtkTreePath *path = NULL;
	GtkCellRendererPixbuf *renderer = NULL;
	GtkWidget *scroll = NULL;
	GtkWidget *vbox = NULL;
	GtkWidget *label = NULL;
	g_return_val_if_fail (self != NULL, NULL);
	g_return_val_if_fail (tree_view != NULL, NULL);

	parent_win = gsampler_main_window_get_instance ();
	dialog = gtk_dialog_new_with_buttons (NULL, (GtkWindow*) parent_win, 
	                                      GTK_DIALOG_DESTROY_WITH_PARENT,
	                                      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
	                                      GTK_STOCK_APPLY, GTK_RESPONSE_APPLY, NULL);
	content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
	vbox = gtk_vbox_new (FALSE, 0);
	db_model = gtk_tree_view_get_model (self->priv->db_tree);
	gtk_tree_view_set_model ((GtkTreeView*) tree_view, db_model);
	renderer = (GtkCellRendererPixbuf*) gtk_cell_renderer_pixbuf_new ();
	g_object_set (renderer, "icon-name", "folder", NULL);
	gtk_tree_view_insert_column_with_attributes ((GtkTreeView*) tree_view, -1, NULL,
	                                             (GtkCellRenderer*) renderer, 
	                                             "pixbuf", 0, NULL, NULL);
	gtk_tree_view_insert_column_with_attributes ((GtkTreeView*) tree_view, -1, NULL,
	                                             (GtkCellRenderer*) gtk_cell_renderer_text_new (),
	                                             "text", 1, NULL, NULL);
	gtk_tree_view_set_headers_visible ((GtkTreeView*) tree_view, FALSE);
	scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_set_size_request (scroll, 300, 400);
	gtk_scrolled_window_set_policy ((GtkScrolledWindow*) scroll, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add ((GtkContainer*) scroll, tree_view);
	label = gtk_label_new (_("Select the destination dir :"));
	gtk_box_pack_start ((GtkBox*) vbox, label, TRUE, FALSE, 0);
	gtk_box_pack_start ((GtkBox*) vbox, scroll, TRUE, FALSE, 0);
	gtk_container_add (GTK_CONTAINER (content_area), vbox);
	gtk_widget_show_all (dialog);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_APPLY);
	db_model = gtk_tree_view_get_model ((GtkTreeView*) tree_view);
	gtk_tree_model_get_iter_first (db_model, &iter);
	path = gtk_tree_model_get_path (db_model, &iter);
	if (path != NULL)
	{
		gtk_tree_view_expand_row ((GtkTreeView*) tree_view, path, FALSE);
	}
	g_object_unref (parent_win);
	return dialog;
}

static gpointer
gsampler_data_base_on_move_thread (gpointer data)
{
	GsamplerClient *client = NULL;
	gboolean ret = FALSE;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (self != NULL, NULL);

	client = gsampler_client_get_instance ();
	if (gsampler_data_base_check_selected_is_dir (self))
	{
		if (g_strcmp0 (self->priv->move_action, "Move") == 0)
			ret = gsampler_client_db_move_dir (client, self->priv->move_inst, self->priv->move_dst);							
		else if (g_strcmp0 (self->priv->move_action, "Copy") == 0)
			ret = gsampler_client_db_copy_dir (client, self->priv->move_inst, self->priv->move_dst);
	} else
	{
		if (g_strcmp0 (self->priv->move_action, "Move") == 0)
			ret = gsampler_client_db_move_instrument (client, self->priv->move_inst, self->priv->move_dst);
		else if (g_strcmp0 (self->priv->move_action, "Copy") == 0)
			ret = gsampler_client_db_copy_instrument (client, self->priv->move_inst, self->priv->move_dst);
	}
	if (!ret)
		g_warning ("file %s: line %d: error: can't move/copy instrument/dir", __FILE__, __LINE__);
	gsampler_data_base_update_instrument_treeview (self);
	g_object_unref (client);
	gsampler_main_window_set_wait_cursor (FALSE);
	return NULL;
}

static void
gsampler_data_base_on_move (GtkMenuItem* item, gpointer data) 
{
	GtkWidget *dialog = NULL;
	GtkWidget *dir_view = NULL;
	GtkTreeModel *model = NULL;
	GtkTreeSelection *select = NULL;
	GtkTreeIter iter = {0};
	gchar *inst = NULL;
	gchar *dst_dir = NULL;
	GError *error = NULL;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);

	dir_view = gtk_tree_view_new (); 
	dialog = gsampler_data_base_tree_dialog (self, dir_view);
	gint result = gtk_dialog_run (GTK_DIALOG (dialog));
	if (result == GTK_RESPONSE_APPLY)
	{
		select = gtk_tree_view_get_selection ((GtkTreeView*) dir_view);
		if (gtk_tree_selection_get_selected (select, &model, &iter)) {
			dst_dir = gsampler_data_base_get_full_path (self, model, &iter);
			if (dst_dir != NULL)
			{
				inst = gsampler_data_base_get_inst_path (self);
				if (inst != NULL)
				{
					self->priv->move_inst = g_strdup (inst);
					self->priv->move_dst = g_strdup (dst_dir);
					self->priv->move_action = (gchar*) gtk_menu_item_get_label (item);
					gsampler_main_window_set_wait_cursor (TRUE);
					g_thread_create (gsampler_data_base_on_move_thread, self, FALSE, &error);
					if (error != NULL)
					{
						g_warning ("file %s: line %d: Can't create thread : %s\n",
								   __FILE__, __LINE__, error->message);
						g_error_free (error);
					}
					
					g_free (inst);
				}
				g_free (dst_dir);
			}
		}
	}
	gtk_widget_destroy (dialog);
}

static void 
gsampler_data_base_on_rename (GtkMenuItem* item, gpointer data) 
{
	GtkWidget *dialog = NULL;
	GtkWidget *content_area = NULL;
	GtkWidget *hbox = NULL;
	GtkWidget *label = NULL;
	GtkWidget *entry = NULL;
	GsamplerClient *client = NULL;
	gchar* dir = NULL;
	gboolean ret = FALSE;
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);

	dialog = gtk_dialog_new_with_buttons (_("Rename Instrument"), NULL, 
	                                      GTK_DIALOG_DESTROY_WITH_PARENT,
	                                      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
	                                      GTK_STOCK_APPLY, GTK_RESPONSE_APPLY, NULL);
	content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
	hbox = gtk_hbox_new (TRUE, 0);
	label = gtk_label_new (_("New name : "));
	entry = gtk_entry_new ();
	gtk_box_pack_start ((GtkBox*) hbox, label, TRUE, FALSE, 0);
	gtk_box_pack_start ((GtkBox*) hbox, entry, TRUE, FALSE, 0);
	gtk_container_add (GTK_CONTAINER (content_area), hbox);
	gtk_widget_show_all (dialog);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_APPLY);
	gint result = gtk_dialog_run (GTK_DIALOG (dialog));
	if (result == GTK_RESPONSE_APPLY)
	{
		gchar *new_name = (gchar*) gtk_entry_get_text ((GtkEntry*) entry);
		if (new_name != NULL)
		{
			dir = gsampler_data_base_get_inst_path (self);
			if (dir != NULL)
			{
				client = gsampler_client_get_instance ();
				if (gsampler_data_base_check_selected_is_dir (self))
					ret = gsampler_client_db_rename_dir (client, dir, new_name);
				else
					ret = gsampler_client_db_rename_instrument (client, dir, new_name);
				if (!ret)
					g_warning ("file %s: line %d: error: can't rename dir/instrument", __FILE__, __LINE__);
				gsampler_data_base_update_instrument_treeview (self);
				g_object_unref (client);
				g_free (dir);
			}
		}
	}
	gtk_widget_destroy (dialog);
}

static void 
gsampler_data_base_on_delete (GtkMenuItem* item, gpointer data) 
{
	GsamplerClient *client = NULL;
	gchar* dir = NULL;
	gboolean ret = FALSE;
	GsamplerDataBase *self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);

	dir = gsampler_data_base_get_inst_path (self);
	if (dir != NULL)
	{
		client = gsampler_client_get_instance ();
		if (gsampler_data_base_check_selected_is_dir (self))
			ret = gsampler_client_db_remove_dir (client, dir);
		else
			ret = gsampler_client_db_remove_instrument (client, dir);
		if (!ret)
			g_warning ("file %s: line %d: error: can't remove instrument", __FILE__, __LINE__);

		gsampler_data_base_update_instrument_treeview (self);
		g_object_unref (client);
		g_free (dir);
	}
}

static void
gsampler_data_base_on_inst_prop (GtkMenuItem* item, gpointer data)
{
	GsamplerClient *client = NULL;
	gchar* inst = NULL;
	GsamplerDbInstInfo *inst_info = NULL;
	GsamplerDataBase *self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);

	inst = gsampler_data_base_get_inst_path (self);
	if (inst != NULL)
	{
		client = gsampler_client_get_instance ();
		inst_info = gsampler_client_db_get_instrument_info (client, inst);
		inst_info->name = gsampler_data_base_get_inst_name (self);
		inst_info->db_path = gsampler_data_base_get_db_path (self);
		GtkDialog *dialog = (GtkDialog*) gsampler_database_instinfo_new (inst_info, NULL);
		gint result = gtk_dialog_run (dialog);
		if (result == GTK_RESPONSE_CLOSE)
			gtk_widget_destroy ((GtkWidget*) dialog);

		gsampler_db_inst_info_free (inst_info);
		g_object_unref (client);
		g_free (inst);
	}
}

static void 
gsampler_data_base_do_popup_dir (GsamplerDataBase* self, GdkEventButton* event) 
{
	guint button = 0U;
	guint32 event_time = 0U;
	GtkMenu* new_menu = NULL;
	GtkMenuItem* new_item = NULL;
	GtkWidget *separator = NULL;
	GtkTreeSelection* select = NULL;
	g_return_if_fail (self != NULL);

	select = gtk_tree_view_get_selection (self->priv->inst_tree);
	if (!gtk_tree_selection_get_selected (select, NULL, &self->priv->current_iter)) 
	{
		return;
	}
	if (event != NULL) 
	{
		button = event->button;
		event_time = event->time;
	} else 
	{
		button = 0;
		event_time = gtk_get_current_event_time ();
	}

	new_menu = (GtkMenu*) gtk_menu_new ();

	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Move"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_move, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Copy"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_move, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Rename"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_rename, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	separator = gtk_separator_menu_item_new ();
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, separator);
	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Delete"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_delete, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	g_object_set (new_menu, "attach-widget", (GtkWidget*) self->priv->inst_tree, NULL);	
	gtk_widget_show_all ((GtkWidget*) new_menu);
	gtk_menu_popup (new_menu, NULL, NULL, NULL, NULL, button, event_time);
	gtk_tree_selection_select_iter (select, &self->priv->current_iter);
	
}

static void 
gsampler_data_base_do_popup_inst (GsamplerDataBase* self, GdkEventButton* event) 
{
	guint button = 0U;
	guint32 event_time = 0U;
	GtkTreeModel* model = NULL;
	GtkTreeSelection* select = NULL;
	GtkMenu* new_menu = NULL;
	GtkMenu* sub_menu = NULL;
	GtkMenuItem* new_item = NULL;
	GtkWidget *separator = NULL;
	GsamplerClient* client = NULL;
	gchar** channels = NULL;
	gint channels_length = 0;
	g_return_if_fail (self != NULL);

	select = gtk_tree_view_get_selection (self->priv->inst_tree);
	if (!gtk_tree_selection_get_selected (select, &model, &self->priv->current_iter)) 
	{
		return;
	}
	if (event != NULL) 
	{
		button = event->button;
		event_time = event->time;
	} else 
	{
		button = 0;
		event_time = gtk_get_current_event_time ();
	}
	new_menu = (GtkMenu*) gtk_menu_new ();

	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Move"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_move, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);

	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Copy"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_move, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	
	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Rename"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_rename, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);

	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Delete"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_delete, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);

	separator = gtk_separator_menu_item_new ();
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, separator);
	
	sub_menu = (GtkMenu*) gtk_menu_new ();
	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Load Instrument"));
	gtk_menu_item_set_submenu (new_item, (GtkWidget*) sub_menu);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);

	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("On new Channel"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_load_on_new, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) sub_menu, (GtkWidget*) new_item);
	client = gsampler_client_get_instance ();
	channels = gsampler_client_get_available_channel (client);		
	if (channels != NULL && g_strcmp0 (channels[0], "NONE") != 0) 
	{
		separator = gtk_separator_menu_item_new ();
		gtk_menu_shell_append ((GtkMenuShell*) sub_menu, separator);
		channels_length = g_strv_length (channels);
		int channel_it;
		for (channel_it = 0; channel_it < channels_length; channel_it = channel_it + 1) 
		{
			gchar* channel;
			channel = g_strdup (channels[channel_it]);
			new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (g_strconcat (_("Channel "), 
			                                                                    channel, NULL));
			g_signal_connect_object (new_item, "activate", 
			                         (GCallback) gsampler_data_base_load_on_chan, self, 0);
			gtk_menu_shell_append ((GtkMenuShell*) sub_menu, (GtkWidget*) new_item);
			g_free (channel);
		}
	}

	separator = gtk_separator_menu_item_new ();
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, separator);

	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Proprieties"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_data_base_on_inst_prop, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	
	g_object_set (new_menu, "attach-widget", (GtkWidget*) self->priv->inst_tree, NULL);	
	gtk_widget_show_all ((GtkWidget*) new_menu);
	gtk_menu_popup (new_menu, NULL, NULL, NULL, NULL, button, event_time);
	gtk_tree_selection_select_iter (select, &self->priv->current_iter);
	g_strfreev (channels);
	g_object_unref (client);
}

static gboolean 
gsampler_data_base_on_inst_button_press_event (GtkWidget* sender, 
                                               GdkEventButton* event, gpointer data) 
{
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (GSAMPLER_IS_DATA_BASE (self), FALSE);
	
	if (event->type == GDK_BUTTON_PRESS && event->button == 3) 
	{
		if (gsampler_data_base_check_selected_is_dir (self))
			gsampler_data_base_do_popup_dir (self, event);
		else
			gsampler_data_base_do_popup_inst (self, event);
			
		return TRUE;
	} else
	{
		return FALSE;
	}
}

static gboolean 
gsampler_data_base_on_popup_inst (GtkWidget* sender, gpointer data) 
{
	GsamplerDataBase* self = GSAMPLER_DATA_BASE (data);
	g_return_val_if_fail (GSAMPLER_IS_DATA_BASE (self), FALSE);

	if (gsampler_data_base_check_selected_is_dir (self))
		gsampler_data_base_do_popup_dir (self, NULL);
	else
		gsampler_data_base_do_popup_inst (self, NULL);
	return TRUE;
}

static void
gsampler_data_base_on_inst_row_activated (GtkTreeView *tree_view, GtkTreePath *path,
                                          GtkTreeViewColumn *column, gpointer data)
{
	GtkTreeSelection *select = NULL;
	GtkTreePath *db_path = NULL;
	gchar *base_name = NULL;
	GsamplerDataBase *self = GSAMPLER_DATA_BASE (data);
	g_return_if_fail (self != NULL);

	if (gsampler_data_base_check_selected_is_dir (self))
	{
		base_name = gsampler_data_base_get_inst_path (self);
		if (base_name != NULL)
		{
			db_path = g_hash_table_lookup (gsampler_data_base_path_list, base_name);
			if (db_path != NULL)
			{
				gtk_tree_view_expand_to_path (self->priv->db_tree, db_path);
				select = gtk_tree_view_get_selection (self->priv->db_tree);
				gtk_tree_selection_select_path (select, db_path);
			}	
			g_free (base_name);
		}
	}
}

static void
gsampler_data_base_inst_tree_data_func (GtkTreeViewColumn *col, GtkCellRenderer *renderer,
                                        GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean dir = FALSE;
	gtk_tree_model_get(model, iter, 2, &dir, -1);
	if (dir)
	{
		g_object_set (renderer, "icon-name", "folder", NULL);
	} else
	{
		g_object_set (renderer, "icon-name", "audio-x-generic", NULL);
	}
}

static void
gsampler_data_base_dir_tree_data_func (GtkTreeViewColumn *col, GtkCellRenderer *renderer,
                                        GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gchar *dir = NULL;
	gtk_tree_model_get(model, iter, 1, &dir, -1);
	if (g_strcmp0 (dir, "/") == 0)
	{
		gchar *load_file = NULL;
		GError *error = NULL;
		GdkPixbuf *pixbuf = NULL;
		load_file = gsampler_utils_get_data_file (DB_ICON);
		if (load_file != NULL)
		{
			pixbuf = gdk_pixbuf_new_from_file_at_scale (load_file, 20, 20, TRUE, &error);
			if (error != NULL) 
			{
				g_warning ("file %s: line %d: error: %s", __FILE__, __LINE__, error->message);
				g_error_free (error);
			} else 
			{
				g_object_set (renderer, "pixbuf", pixbuf, NULL);
			}
			g_free (load_file);
		}
	} else
	{
		g_object_set (renderer, "icon-name", "folder", NULL);
	}
	g_free (dir);
}

static GObject * 
gsampler_data_base_constructor (GType type, guint n_construct_properties, 
                                 GObjectConstructParam * construct_properties) 
{
	GObject * obj;
	GsamplerDataBase * self = NULL;
	GtkToolbar* db_tool = NULL;
	GtkScrolledWindow* scroll1 = NULL;
	GtkScrolledWindow* scroll2 = NULL;
	GtkImage* add_image = NULL;
	//GtkImage* del_image = NULL;
	GtkImage* find_image = NULL;
	GtkImage* clear_image = NULL;
	GtkImage* sup_image = NULL;
	GtkImage* refresh_image = NULL;
	GtkToolButton* add_button = NULL;
	//GtkToolButton* del_button = NULL;
	GtkToolButton* find_button = NULL;
	GtkToolButton* clear_button = NULL;
	GtkToolButton* sup_button = NULL;
	GtkToolButton* refresh_button = NULL;
	GtkTreeStore* store = NULL;
	GtkCellRendererPixbuf* renderer;
	GtkTreeViewColumn *col = NULL;
	GtkListStore* liste = NULL;
	GtkTreeSelection* select1 = NULL;
									 
	obj = G_OBJECT_CLASS (gsampler_data_base_parent_class)->constructor (type, n_construct_properties, construct_properties);
	self = GSAMPLER_DATA_BASE (obj);
	gtk_orientable_set_orientation ((GtkOrientable*) self, GTK_ORIENTATION_VERTICAL);

	db_tool = (GtkToolbar*) gtk_toolbar_new ();
	add_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_ADD, 
	                                                  GTK_ICON_SIZE_SMALL_TOOLBAR);
	/*del_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_DELETE, 
	                                                  GTK_ICON_SIZE_SMALL_TOOLBAR);*/
	find_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_FIND, 
	                                                   GTK_ICON_SIZE_SMALL_TOOLBAR);
	clear_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_CLEAR, 
	                                                    GTK_ICON_SIZE_SMALL_TOOLBAR);
	sup_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_DELETE, 
	                                                    GTK_ICON_SIZE_SMALL_TOOLBAR);
	refresh_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_REFRESH, 
	                                                      GTK_ICON_SIZE_SMALL_TOOLBAR);
	add_button = (GtkToolButton*) gtk_tool_button_new ((GtkWidget*) add_image, 
	                                                   _("Add sample"));
	gtk_tool_item_set_tooltip_text ((GtkToolItem*) add_button, 
	                                _("Add sample to the database"));
	/*del_button = (GtkToolButton*) gtk_tool_button_new ((GtkWidget*) del_image, 
	                                                   _("Delete sample"));*/
	/*gtk_tool_item_set_tooltip_text ((GtkToolItem*) del_button, 
	                                _("Delete sample in the database"));*/
	find_button = (GtkToolButton*) gtk_tool_button_new ((GtkWidget*) find_image, 
	                                                    _("Find inst"));
	gtk_tool_item_set_tooltip_text ((GtkToolItem*) find_button, 
	                                _("Find instrument in the database"));
	clear_button = (GtkToolButton*) gtk_tool_button_new ((GtkWidget*) clear_image, 
	                                                     _("clear the db"));
	gtk_tool_item_set_tooltip_text ((GtkToolItem*) clear_button,
	                                _("Clear the database"));
	sup_button = (GtkToolButton*) gtk_tool_button_new ((GtkWidget*) sup_image, 
	                                                     _("format the db"));
	gtk_tool_item_set_tooltip_text ((GtkToolItem*) sup_button,
	                                _("Format the database"));
	refresh_button = (GtkToolButton*) gtk_tool_button_new ((GtkWidget*) refresh_image,
	                                                       _("Refresh the db"));
	gtk_tool_item_set_tooltip_text ((GtkToolItem*) refresh_button,
	                                _("Refresh the database view"));	
	gtk_container_add ((GtkContainer*) db_tool, (GtkWidget*) add_button);
	//gtk_container_add ((GtkContainer*) db_tool, (GtkWidget*) del_button);
	gtk_container_add ((GtkContainer*) db_tool, (GtkWidget*) find_button);
	gtk_container_add ((GtkContainer*) db_tool, (GtkWidget*) clear_button);
	gtk_container_add ((GtkContainer*) db_tool, (GtkWidget*) sup_button);
	gtk_container_add ((GtkContainer*) db_tool, (GtkWidget*) refresh_button);
	gtk_toolbar_set_icon_size (db_tool, GTK_ICON_SIZE_MENU);
	
	store = gtk_tree_store_new (2, GDK_TYPE_PIXBUF, G_TYPE_STRING);
	self->priv->db_tree = (GtkTreeView*) gtk_tree_view_new_with_model ((GtkTreeModel*) store);
	renderer = (GtkCellRendererPixbuf*) gtk_cell_renderer_pixbuf_new ();
	//g_object_set (renderer, "icon-name", "folder", NULL);
	gtk_tree_view_insert_column_with_attributes (self->priv->db_tree, -1, NULL,
	                                             (GtkCellRenderer*) renderer, 
	                                             "pixbuf", 0, NULL, NULL);
	col = gtk_tree_view_get_column (self->priv->db_tree, 0);
	gtk_tree_view_column_set_cell_data_func(col, (GtkCellRenderer*) renderer, gsampler_data_base_dir_tree_data_func, NULL, NULL);
	gtk_tree_view_insert_column_with_attributes (self->priv->db_tree, -1, NULL,
	                                             (GtkCellRenderer*) gtk_cell_renderer_text_new (),
	                                             "text", 1, NULL, NULL);
	gtk_tree_view_set_headers_visible (self->priv->db_tree, FALSE);
	scroll1 = (GtkScrolledWindow*) gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (scroll1, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	liste = gtk_list_store_new (3, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_BOOLEAN);
	self->priv->inst_tree = (GtkTreeView*) gtk_tree_view_new_with_model ((GtkTreeModel*) liste);
	renderer = (GtkCellRendererPixbuf*) gtk_cell_renderer_pixbuf_new ();
	gtk_tree_view_insert_column_with_attributes (self->priv->inst_tree, -1, NULL, 
	                                             (GtkCellRenderer*) renderer, 
	                                             "pixbuf", 0, NULL, NULL);
	col = gtk_tree_view_get_column (self->priv->inst_tree, 0);
	gtk_tree_view_column_set_cell_data_func(col, (GtkCellRenderer*) renderer, gsampler_data_base_inst_tree_data_func, NULL, NULL);
	gtk_tree_view_insert_column_with_attributes (self->priv->inst_tree, -1, "Name", 
	                                             (GtkCellRenderer*) gtk_cell_renderer_text_new (), 
	                                             "text", 1, NULL, NULL);
	gtk_tree_view_set_hover_selection (self->priv->inst_tree, TRUE);
	
	scroll2 = (GtkScrolledWindow*) gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (scroll2, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	
	gtk_container_add ((GtkContainer*) scroll1, (GtkWidget*) self->priv->db_tree);
	gtk_container_add ((GtkContainer*) scroll2, (GtkWidget*) self->priv->inst_tree);
	gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) db_tool, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) scroll1, TRUE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) scroll2, TRUE, TRUE, 0);
	gsampler_data_base_path_list = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, 
	                                                      (GDestroyNotify) gtk_tree_path_free);
	
	select1 = gtk_tree_view_get_selection (self->priv->db_tree);
	g_signal_connect_object (select1, "changed", 
	                         (GCallback) gsampler_data_base_on_folder_changed, self, 0);
	g_signal_connect_object (self->priv->inst_tree, "row-activated", 
	                         (GCallback) gsampler_data_base_on_inst_row_activated, self, 0);
	g_signal_connect_object ((GtkWidget*) self->priv->inst_tree, "popup-menu", 
	                         (GCallback) gsampler_data_base_on_popup_inst, self, 0);
	g_signal_connect_object ((GtkWidget*) self->priv->inst_tree, "button-press-event", 
	                         (GCallback) gsampler_data_base_on_inst_button_press_event, self, 0);
	g_signal_connect_object (add_button, "clicked", 
	                         (GCallback) gsampler_data_base_on_add_db_clicked, self, 0);
	g_signal_connect_object (find_button, "clicked", 
	                         (GCallback) gsampler_data_base_on_find_db_clicked, self, 0);
	g_signal_connect_object (clear_button, "clicked", 
	                         (GCallback) gsampler_data_base_on_clear_db_clicked, self, 0);
	g_signal_connect_object (sup_button, "clicked", 
	                         (GCallback) gsampler_data_base_on_format_db_clicked, self, 0);
	g_signal_connect_object (refresh_button, "clicked", 
	                         (GCallback) gsampler_data_base_on_refresh_db_clicked, self, 0);
	gtk_widget_show_all ((GtkWidget*) self);
	
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);
	g_object_unref (liste);
	g_object_unref (store);
	return obj;
}

static void 
gsampler_data_base_init (GsamplerDataBase * self) 
{
	self->priv = GSAMPLER_DATA_BASE_GET_PRIVATE (self);
	self->priv->new_rep = NULL;
	self->priv->move_inst = NULL;
	self->priv->move_action = NULL;
	self->priv->move_dst = NULL;
	self->priv->add_job_id = 0;
}

static void 
gsampler_data_base_finalize (GObject* obj) 
{
	GsamplerDataBase * self;
	self = GSAMPLER_DATA_BASE (obj);
	g_free (self->priv->new_rep);
	g_free (self->priv->move_inst);
	g_free (self->priv->move_action);
	g_free (self->priv->move_dst);
	G_OBJECT_CLASS (gsampler_data_base_parent_class)->finalize (obj);
}

static void 
gsampler_data_base_class_init (GsamplerDataBaseClass * klass) 
{
	gsampler_data_base_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GsamplerDataBasePrivate));
	G_OBJECT_CLASS (klass)->finalize = gsampler_data_base_finalize;
	G_OBJECT_CLASS (klass)->constructor = gsampler_data_base_constructor;
}

GsamplerDataBase* 
gsampler_data_base_new (void) 
{
	return g_object_new (GSAMPLER_TYPE_DATA_BASE, NULL);
}

void
gsampler_data_base_active (GsamplerDataBase* self) 
{
	g_return_if_fail (GSAMPLER_IS_DATA_BASE (self));

	gdk_threads_add_idle_full (G_PRIORITY_DEFAULT_IDLE, gsampler_data_base_initialyse,
	                 		   g_object_ref (self), g_object_unref);	
}

void 
gsampler_data_base_inactive (GsamplerDataBase* self) 
{
	GsamplerClient* client;
	GtkListStore* liste = NULL;
	GtkTreeStore* store = NULL;
	g_return_if_fail (GSAMPLER_IS_DATA_BASE (self));
	
	liste = (GtkListStore*) gtk_tree_view_get_model (self->priv->inst_tree);
	gtk_list_store_clear (liste);
	store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->db_tree);
	gtk_tree_store_clear (store);
	g_hash_table_remove_all (gsampler_data_base_path_list);
	client = gsampler_client_get_instance ();
	g_signal_handlers_disconnect_by_func (client, gsampler_data_base_on_db_path_count, self);
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);
	g_object_unref (client);	
}